import bs4
from bs4 import BeautifulSoup
from yattag import Doc

from pacu.app.report.htmlbase import HtmlBase


class HtmlElement(HtmlBase):
    """
    This class contains a custom HTML element.
    """

    def __init__(self, tag, text=None, attributes=None):
        """
        Initializes an HTML element.
        :param tag: Tag
        :param attributes: HTML attributes
        :param text: Text
        """
        super(HtmlElement, self).__init__()
        self._tag_name = tag
        self._attributes = attributes
        self._tag_text = text

    @property
    def text(self):
        """
        Returns the HTML element as plain text.
        :return: Text
        """
        return self._tag_text

    def _has_nested_content(self):
        """
        Returns true if this HTML element has nested content.
        :return: True if there is nested content
        """
        return len(self._doc.getvalue()) != 0

    # noinspection PyArgumentList
    def to_html(self):
        """
        Converts this element to HTML code. A novel Doc() instance is created in order to nest the content of this
        elements Doc() inside the tag associated with this HtmlElement.
        :return: HTML code
        """
        if self._attributes is None:
            self._attributes = []
        doc, tag, text = Doc().tagtext()
        if not self._has_nested_content() and self._tag_text is None:
            # If there is no content / text in the tag, a self closing tag (stag) is created
            doc.stag(self._tag_name, *self._attributes)
        else:
            # Otherwise the content / text is added inside of the tag
            with tag(self._tag_name, *self._attributes):
                if self._tag_text is not None:
                    html = BeautifulSoup(str(self._tag_text), 'html.parser')
                    for part in html.contents:
                        if isinstance(part, bs4.element.Tag):
                            doc.asis(f'<{part.name}>')
                            doc.text(part.text)
                            doc.asis(f'</{part.name}>')
                        else:
                            # noinspection PyTypeChecker
                            doc.text(part)
                if self._has_nested_content():
                    doc.asis(self._doc.getvalue())
        return doc.getvalue()



from importlib.resources import files
from pathlib import Path
from typing import Optional, Dict

from pacu.app.command import Command
from pacu.app.utils.loggingutils import logger

OPTIONS = {
    'branch_swap_filter': ['None', 'Very Weak', 'Weak', 'Moderate', 'Strong', 'Very Strong'],
    'missing_data_treatment': ['Complete deletion', 'Use all sites', 'Partial deletion'],
}

SUBSTITUTION_MODELS = {
    'JC': 'Jukes-Cantor model',
    'K2': 'Kimura 2-parameter model',
    'T92': 'Tamura 3-parameter model',
    'HKY': 'Hasegawa-Kishino-Yano model',
    'TN93': 'Tamura-Nei model',
    'GTR': 'General Time Reversible model'
}

HEURISTIC_METHODS = {
    'NNI': 'Nearest-Neighbor-Interchange (NNI)',
    'SPR3': 'Subtree-Pruning-Regrafting - Fast (SPR level 3)',
    'SPR5': 'Subtree-Pruning-Regrafting - Extensive (SPR level 5)'
}

INITIAL_TREE = {
    'NJ_BioNJ': 'Make initial tree automatically (Default - NJ/BioNJ)',
    'Max_Parsimony': 'Make initial tree automatically (Maximum Parsimony)',
    'NJ': 'Make initial tree automatically (Neighbor Joining)',
    'BioNJ': 'Make initial tree automatically (BioNJ)'
}

RATES_AMONG_SITES = {
    'U': 'Uniform Rates',
    'G': 'Gamma Distributed (G)',
    'I': 'Has Invariant Sites (I)',
    'G+I': 'Gamma Distributed With Invariant Sites (G+I)'
}


def __create_config_file_model_selection(
        dir_: Path, branch_swap_filter: str, missing_data_treatment: str, site_cov_cutoff: Optional[int],
        threads: int = 4) -> Path:
    """
    Generates the config file for the model selection.
    :param dir_: Working directory
    :param missing_data_treatment: Missing data treatment option
    :param site_cov_cutoff: Site coverage cut-off threshold
    :param threads: Number of threads to use
    :return: Path to config file
    """
    # Parse the template
    path_template = files('pacu').joinpath('resources/mega/model_sel_ml_nucleotide_template.mao')
    with path_template.open() as handle:
        template = handle.read()
    logger.info(f'Template parsed')

    # Create filled-in template
    path_out = dir_ / 'config-model_selection.mao'
    with path_out.open('w') as handle:
        handle.write(template.format(
            branch_swap_filter=branch_swap_filter,
            missing_data_treatment=missing_data_treatment,
            site_coverage_cutoff=site_cov_cutoff if site_cov_cutoff is not None else 'Not Applicable',
            threads=threads
        ))
    return path_out


def run_model_selection(path_fasta: Path, path_out: Path, dir_: Path, branch_swap_filter: str,
                        missing_data_treatment: str, site_cov_cutoff: Optional[int], threads: int = 4) -> None:
    """
    Runs model selection on the given SNP matrix.
    :param path_fasta: Input SNP matrix FASTA file
    :param path_out: Output path
    :param dir_: Working directory
    :param branch_swap_filter: Branch swap filter option
    :param missing_data_treatment: Missing data treatment option
    :param site_cov_cutoff: Site coverage cut-off
    :param threads: Number of threads to use
    :return: None
    """
    # Create configuration file
    path_config = __create_config_file_model_selection(
        dir_, branch_swap_filter, missing_data_treatment, site_cov_cutoff, threads)

    # Run MEGA model selection
    command = Command(' '.join([
        'megacc', f'-d {path_fasta}', f'-a {path_config}', f'-o {path_out}'
    ]))
    command.run(path_out.parent)
    if not command.exit_code == 0:
        raise RuntimeError(f'Error running model selection: {command.stderr}')


def parse_model_selection_csv(path_csv: Path) -> Dict:
    """
    Parses the model selection output CSV file.
    :return: Information for the best selected model
    """
    model_info = {}
    with path_csv.open() as handle:
        first_line = handle.readlines()[1]
        # The line has the following structure: K2+G+I, ...
        # The model is the first part (K2), +G means Gamma categories per site, +I means invariant sites
        complete_model = first_line.split(',')[0]
        model_info['model'] = first_line.split(',')[0].split('+')[0]
        model_info['model_full'] = SUBSTITUTION_MODELS[model_info['model']]

        # Rates
        complete_rates = '+'.join(complete_model.split('+')[1:])
        model_info['rates_among_sites'] = 'U' if complete_rates == '' else complete_rates
        model_info['rates_among_sites_full'] = RATES_AMONG_SITES[model_info['rates_among_sites']]
    return model_info


def __create_config_file_tree_construction(
        dir_: Path, model: str, branch_swap_filter: str, missing_data_treatment: str, site_cov_cutoff: Optional[int],
        bootstrap_replicates: int, heuristic_method: str, initial_tree: str, rates_among_sites: str,
        gamma_categories: int = 5, threads: int = 4) -> Path:
    """
    Generates the config file for the model selection.
    :param dir_: Working directory
    :param model: Nucleotide substitution model
    :param missing_data_treatment: Missing data treatment option
    :param site_cov_cutoff: Site coverage cut-off threshold
    :param bootstrap_replicates: Number of bootstrap replications
    :param heuristic_method: Heuristic method
    :param initial_tree: Initial tree to start from
    :param rates_among_sites: Rates among sites
    :param gamma_categories: Number of Gamma categories
    :param threads: Number of threads to use
    :return: Path to config file
    """
    # Parse the template
    path_template = files('pacu').joinpath('resources/mega/infer_ml_nucleotide_template.mao')
    with path_template.open() as handle:
        template = handle.read()
    logger.info(f'Template parsed')

    # Create filled-in template
    path_out = dir_ / 'config-tree_building.mao'
    with path_out.open('w') as handle:
        handle.write(template.format(
            bootstrap_replications=bootstrap_replicates,
            model=model,
            heuristic_method=HEURISTIC_METHODS[heuristic_method],
            initial_tree=INITIAL_TREE[initial_tree],
            rates_among_sites=RATES_AMONG_SITES[rates_among_sites],
            branch_swap_filter=branch_swap_filter,
            gamma_categories=gamma_categories,
            missing_data_treatment=missing_data_treatment,
            site_coverage_cutoff=site_cov_cutoff,
            threads=threads
        ))
    return path_out


def run_tree_building(path_fasta: Path, path_csv: Path, path_out: Path, dir_: Path, branch_swap_filter: str,
                      missing_data_treatment: str, site_cov_cutoff: Optional[int], bootstrap_replicates: int,
                      heuristic_method: str, initial_tree: str, gamma_categories: int = 5,
                      threads: int = 4) -> None:
    """
    Runs the ML tree reconstruction using MEGA.
    :return: None
    """
    # Parse model
    data_model_selection = parse_model_selection_csv(path_csv)

    # Create configuration file
    path_config = __create_config_file_tree_construction(
        dir_, data_model_selection['model_full'], branch_swap_filter, missing_data_treatment, site_cov_cutoff,
        bootstrap_replicates, heuristic_method, initial_tree, data_model_selection['rates_among_sites'],
        gamma_categories, threads)

    # Run MEGA model selection
    command = Command(' '.join([
        'megacc', f'-d {path_fasta}', f'-a {path_config}', f'-o {path_out}'
    ]))
    command.run(path_out.parent)
    if not command.exit_code == 0:
        raise RuntimeError(f'Error running tree construction: {command.stderr}')



#!/usr/bin/env python
import argparse
from pathlib import Path
from typing import Tuple, List

from pacu import initialize_logging, logger, PACU
from pacu.app.utils import workflowutils, bamutils


def parse_galaxy_args() -> Tuple[argparse.Namespace, List[str]]:
    """
    Parses the Galaxy arguments.
    :return: Parsed arguments
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('--ref-fasta', required=True, type=Path)
    parser.add_argument('--ref-fasta-name', type=str)
    parser.add_argument('--bam-ilmn', nargs=2, action='append')
    parser.add_argument('--bam-ont', nargs=2, action='append')
    parser.add_argument('--dir-working', default=Path.cwd(), type=Path, help='Working directory')
    return parser.parse_known_args()


def _extract_dataset_name(bam_in: Path, name_orig: str) -> str:
    """
    Extracts the dataset name from the input BAM file.
    :param bam_in: Input BAM file
    :param name_orig: Galaxy name
    :return: Extracted name
    """
    try:
        return bamutils.read_custom_tag(bam_in, 'PACU_name')
    except ValueError:
        logger.debug(f"'PACU_name' not found in BAM header: {name_orig}")
    return workflowutils.sanitize_bam_input(name_orig)


def main() -> None:
    """
    Runs the main script.
    :return: None
    """
    initialize_logging()
    logger.info(f'Running PACU through Galaxy')
    args, unparsed_args = parse_galaxy_args()

    # Create input directory
    dir_in = args.dir_working / 'input'
    dir_in.mkdir(exist_ok=True)

    # Symlink the reference genome
    path_ref_link = dir_in / args.ref_fasta_name
    if not path_ref_link.exists():
        path_ref_link.symlink_to(Path(args.ref_fasta))
    unparsed_args.extend(['--ref-fasta', str(path_ref_link.absolute())])

    # Symlink the BAM files (Illumina)
    if args.bam_ilmn is not None:
        logger.info(f'Creating symlinks for Illumina input ({len(args.bam_ilmn)} datasets)')
        dir_bam = dir_in / 'bam_ilmn'
        dir_bam.mkdir(exist_ok=True)
        for path_bam, galaxy_name in args.bam_ilmn:
            try:
                basename = f'{_extract_dataset_name(Path(path_bam), galaxy_name)}.bam'
                (dir_bam / basename).symlink_to(path_bam)
            except FileExistsError:
                logger.debug(f"Symlink for '{galaxy_name}' already exists")
        unparsed_args.extend(['--ilmn-in', str(dir_bam.absolute())])

    # Symlink the BAM files (ONT)
    if args.bam_ont is not None:
        logger.info(f'Creating symlinks for ONT input ({len(args.bam_ont)} datasets)')
        dir_bam = dir_in / 'bam_ont'
        dir_bam.mkdir(exist_ok=True)
        for path_bam, galaxy_name in args.bam_ont:
            try:
                basename = f'{_extract_dataset_name(Path(path_bam), galaxy_name)}.bam'
                (dir_bam / basename).symlink_to(path_bam)
            except FileExistsError:
                logger.debug(f"Symlink for '{galaxy_name}' already exists")
        unparsed_args.extend(['--ont-in', str(dir_bam.absolute())])

    # Run the main script
    workflow = PACU(unparsed_args)
    workflow.run()


if __name__ == '__main__':
    main()



import datetime
from pathlib import Path
from typing import Any, Dict

import pandas as pd
from matplotlib import pyplot
from upsetplot import plot

from pacu import version
from pacu.app.report.htmlcitation import HtmlCitation
from pacu.app.report.htmlelement import HtmlElement
from pacu.app.report.htmlreportsection import HtmlReportSection
from pacu.app.report.htmltablecell import HtmlTableCell


def create_upsetplot_overlap(data_overlap: pd.DataFrame, png_out: Path) -> None:
    """
    Creates an UpSet plot for the region filtering overlaps.
    :param data_overlap: Overlap data
    :param png_out: Output PNG file
    :return: None
    """
    plot(data_overlap.groupby(by=['phages', 'gubbins', 'depth']).size(), show_counts=True)
    pyplot.savefig(png_out, dpi=300)


def create_analysis_info_section(config_: Dict[str, Any]) -> HtmlReportSection:
    """
    Creates the analysis info section.
    :param config_: Configuration data
    :return: Analysis info section
    """
    section = HtmlReportSection('Analysis info')

    # Analysis info
    section.add_table([
        ['Workflow version:', version.__version__],
        ['Analysis date:', datetime.datetime.now().strftime('%d/%m/%Y - %X')],
        ['Nb. of samples:', str(len(config_['input'].keys()))],
        ['Phylogenetic method:', config_.get('phylogeny_method', 'iqtree')],
    ], table_attributes=[('class', 'information')])

    # Citation disclaimer
    section.add_header('Disclaimer', 2)
    section.add_paragraph('If you use this pipeline for your scientific work, please cite:')
    section.add_html_object(HtmlCitation.parse_from_json('Bogaerts_2023-ont_outbreak'))

    return section


def create_parameter_section(config_: Dict[str, Any]) -> HtmlReportSection:
    """
    Creates the parameter section.
    :param config_: Configuration data
    :return: Section
    """
    section = HtmlReportSection('Parameters')
    section.add_table([
        ['Min. allele frequency:', f"{config_['filters']['min_af']:.2f}"],
        ['Min. depth:', str(config_['filters']['min_depth'])],
        ['Min. SNP quality:', str(config_['filters']['min_qual'])],
    ], table_attributes=[('class', 'information')])
    return section


def __get_colored_cell_depth(value: int) -> HtmlTableCell:
    """
    Returns a colored cell for the mapping output.
    :param value: Depth value
    :return: Table cell
    """
    value_str = f'{value:,}'
    if value >= 20:
        color = 'green'
    elif value < 10:
        color = 'red'
    else:
        color = 'orange'
    return HtmlTableCell(value_str, color=color)


def __get_colored_cell_coverage(value: int) -> HtmlTableCell:
    """
    Returns a colored cell for the mapping output.
    :param value: Coverage value
    :return: Table cell
    """
    value_str = f'{value:.2f}%'
    if value >= 95:
        color = 'green'
    elif value < 90:
        color = 'red'
    else:
        color = 'orange'
    return HtmlTableCell(value_str, color=color)


def create_mapping_section(path_stats: Path, path_ref: Path) -> HtmlReportSection:
    """
    Creates the read mapping section.
    :param path_stats: Path to the TSV stats file
    :param path_ref: Path to reference genome
    :return: Section
    """
    section = HtmlReportSection('Read mapping')
    data_vc = pd.read_table(path_stats)
    header = ['Isolate', 'Median  depth', '% covered']
    section.add_paragraph(f'Reference genome: {path_ref.name}')
    section.add_table([[
        row['key'],
        __get_colored_cell_depth(row['median_depth']),
        __get_colored_cell_coverage(row['perc_covered']),
    ] for row in data_vc.to_dict('records')], header, [('class', 'data')])
    section.add_paragraph(
        """
        Datasets with a median depth between 20x-10x are displayed in orange (warning). Datasets with a median depth 
        <10x are displayed in red and should be re-sequenced. The genome coverage cut-offs are 95% (warning) and 90% 
        (error).
        """
    )
    return section


def create_variant_calling_section(path_stats: Path) -> HtmlReportSection:
    """
    Creates the variant calling section.
    :param path_stats: Path to the TSV stats file
    :return: Section
    """
    section = HtmlReportSection('Variant calling and filtering')
    data_vc = pd.read_table(path_stats)
    columns = [
        {'key': 'key', 'header': 'Isolate'},
        {'key': 'nb_variants', 'header': 'Number of variants', 'fmt': lambda x: f'{x:,}'},
        {'key': 'nb_variants_filtered', 'header': 'Number of variants (Filtered)', 'fmt': lambda x: f'{x:,}'},
        {'key': 'perc_passed', 'header': '% passing filtering', 'fmt': lambda x: f'{x:.2f}'},
    ]
    # noinspection PyCallingNonCallable
    section.add_table([
        [col['fmt'](row[col['key']]) if col.get('fmt') is not None else row[col['key']] for col in columns] for
        row in data_vc.to_dict('records')], [c['header'] for c in columns], [('class', 'data')])
    return section


def create_region_filtering_section(path_stats: Path, path_png: Path) -> HtmlReportSection:
    """
    Creates the section with the region filtering results.
    :param path_stats: Path to TSV stats file
    :param path_png: Visualization of region overlap
    :return: Section
    """
    section = HtmlReportSection('Region filtering')

    # Add stats
    data_regions = pd.read_table(path_stats)
    columns = [
        {'key': 'key', 'header': 'BED file'},
        {'key': 'nb_covered_positions', 'header': 'Positions covered', 'fmt': lambda x: f'{x:,}'},
        {'key': 'perc_ref_covered', 'header': '% covered', 'fmt': lambda x: f'{x:.2f}'},
    ]
    # noinspection PyCallingNonCallable
    section.add_table([
        [col['fmt'](row[col['key']]) if col.get('fmt') is not None else row[col['key']] for col in columns] for
        row in data_regions.to_dict('records')], [c['header'] for c in columns], [('class', 'data')])

    # Add path to image
    relative_path_png = Path(path_png.name)
    section.add_file(path_png, relative_path_png)
    section.add_line_break()
    return section


def create_tree_section(path_nwk: Path, path_png: Path, path_snp_matrix: Path, model_name: str) -> HtmlReportSection:
    """
    Creates the section with the visualization of the tree.
    :param path_nwk: Newick tree
    :param path_png: Tree visualization
    :param path_snp_matrix: SNP matrix
    :param model_name: Name of the model used
    :return: Section
    """
    section = HtmlReportSection('Phylogeny')

    # Add SNP matrix
    relative_path_snp_matrix = Path(path_snp_matrix.name)
    section.add_header('SNP matrix', 4)
    section.add_link_to_file('Download (FASTA)', relative_path_snp_matrix)
    section.add_file(path_snp_matrix, relative_path_snp_matrix)

    # Add visualization
    section.add_header('Tree', 4)
    section.add_paragraph(f'Selected model: {model_name}')
    relative_path_png = Path(path_png.name)
    section.add_html_object(HtmlElement('img', attributes=[('src', str(relative_path_png)), ('border', '1')]))
    section.add_file(path_png, relative_path_png)
    section.add_line_break()

    # Add download link (Newick file)
    relative_path = Path(path_nwk.name)
    section.add_file(path_nwk, relative_path)
    section.add_link_to_file('Download (NWK)', relative_path)
    return section


def create_snp_distances_section(path_distances: Path) -> HtmlReportSection:
    """
    Creates the section with pair-wise SNP distances.
    :param path_distances: Path with pair-wise SNP distances
    :return: Section
    """
    section = HtmlReportSection(f'Pair-wise SNP distances')
    data_snp_dists = pd.read_table(path_distances)
    header = [HtmlElement('th', '')] + [
        HtmlElement('th', col, [('style', 'writing-mode: vertical-lr;')]) for col in data_snp_dists.columns[1:]]
    section.add_table(
        [header] +
        [[record[col] for col in data_snp_dists.columns] for record in data_snp_dists.to_dict('records')],
        None, [('class', 'data')])
    return section


def create_citations_section() -> HtmlReportSection:
    """
    Creates the report section with the citations.
    :return: Section
    """
    section = HtmlReportSection('Citations')
    keys = [
        'Croucher_2015-gubbins',
        'Danecek_2021-bcftools',
        'Li_2018-minimap2',
        'Quinlan_2010-bedtools',
        'Nguyen_2015-iq_tree',
        'Kalyaanamoorthy_2017-iq_tree_model',
        'Kumar_2018-megax'
    ]
    for citation_key in keys:
        section.add_html_object(HtmlCitation.parse_from_json(citation_key))
    return section



from pathlib import Path
from typing import Tuple, List, Union, Optional, Sequence

import abc
import bs4
import yattag
from bs4 import BeautifulSoup
from yattag import SimpleDoc


class HtmlBase(object):
    """
    Base class for HTML objects, contains common methods for classes that implement HTML objects.
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self):
        """
        Initializes an HTML base object.
        """
        self._doc = yattag.SimpleDoc()

    def get_tag(self, tag: str, attributes: List[Tuple[str, str]] = None) -> SimpleDoc.Tag:
        """
        Returns a tag with the given attributes.
        :param tag: Tag
        :param attributes: HTML attributes
        :return: Tag
        """
        if attributes:
            return self._doc.tag(tag, *attributes)
        else:
            return self._doc.tag(tag)

    def add_header(self, text: str, level: int, attributes: List[Tuple[str, str]] = None) -> None:
        """
        Adds a header.
        :param text: Header text
        :param level: Header level (1-6)
        :param attributes: Additional HTML arguments (key-value)
        :return: None
        """
        with self.get_tag('h{}'.format(str(level)), attributes):
            self.add_text(text)

    def add_paragraph(self, text: str, attributes: List[Tuple[str, str]] = None):
        """
        Adds a text paragraph
        :param text: Paragraph text
        :param attributes: Additional HTML arguments (key-value)
        :return: None
        """
        with self.get_tag('p', attributes):
            self.add_text(text)

    def add_line_break(self):
        """
        Adds a line break.
        :return: None
        """
        self._doc.stag('br')

    def to_html(self):
        """
        Converts the HTML object to HTML code.
        :return: HTML code
        """
        return self._doc.getvalue()

    def add_table(self, data: List[Sequence[Union[str, int, 'HtmlBase']]], column_names: List[str] = None,
                  table_attributes: List[Tuple[str, str]] = None) -> None:
        """
        Adds a table.
        :param data: Table data
        :param column_names: Column names
        :param table_attributes: Attributes of the table
        :return: None
        """
        with self.get_tag('table', table_attributes):
            if column_names is not None:
                self._add_table_header(column_names)
            for row in data:
                self._add_table_row(row)
        self.add_line_break()

    def _add_table_header(self, column_names: List[str]) -> None:
        """
        Adds a header to a table.
        :param column_names: Names of the columns
        :return: None
        """
        with self.get_tag('tr'):
            for name in column_names:
                with self.get_tag('th'):
                    self.add_text(name)

    def _add_table_row(self, data: Sequence[Union[str, int]], attributes: Optional[List[Tuple[str, str]]] = None) -> \
            None:
        """
        Adds a row to a table.
        :param data: Row data
        :param attributes: Attributes (optional)
        :return: None
        """
        with self.get_tag('tr', attributes):
            for value in data:
                if isinstance(value, HtmlBase):
                    self._doc.asis(value.to_html())
                else:
                    with self.get_tag('td'):
                        self.add_text(value)

    def add_link_to_file(self, link_text: str, relative_path: Path) -> None:
        """
        Adds a link to a file.
        :param link_text: Text of the link
        :param relative_path: Relative path to the file
        :return: None
        """
        if relative_path is None:
            self._doc.text(link_text)
            with self._doc.tag('em', klass='not-available'):
                self._doc.text('Not available')
        else:
            with self._doc.tag('a', href=str(relative_path)):
                self._doc.text(link_text)
        self.add_line_break()

    def add_error_message(self, message: str) -> None:
        """
        Adds an error message.
        :param message: Message text
        :return: None
        """
        self.add_alert(message, 'error')

    def add_warning_message(self, message: str) -> None:
        """
        Adds a warning message.
        :param message: Message text
        :return: None
        """
        self.add_alert(message, 'warning')

    def add_alert(self, message: str, type_: str) -> None:
        """
        Adds an alert to the report.
        Supported types: 'error', 'warning' and 'info'
        :param message: Message to display
        :param type_: Message type
        :return: None
        """
        with self._doc.tag('div', klass=f'alert-box {type_}'):
            with self._doc.tag('span'):
                self._doc.text(f'{type_}: ')
            self.add_text(message)

    def add_labeled_list(self, rows: List[Sequence[str]], ordered=False,
                         attributes: List[Tuple[str, str]] = None) -> None:
        """
        Adds a labeled list.
        :param rows: List of label-text pairs
        :param ordered: If True an ordered list is added
        :param attributes: Additional HTML arguments (key-value)
        :return: None
        """
        list_tag = 'ol' if ordered else 'ul'
        with self.get_tag(list_tag, attributes):
            for label, text in rows:
                with self._doc.tag('li'):
                    with self._doc.tag('b'):
                        self._doc.text('{}: '.format(label))
                    self.add_text(text)

    def add_html_object(self, input_object: 'HtmlBase') -> None:
        """
        Adds an HTML object.
        :param input_object: Input object
        :return: None
        """
        if not isinstance(input_object, HtmlBase):
            raise ValueError("{} is not an HTML object".format(input_object))
        self._doc.asis(input_object.to_html())

    def add_raw(self, html_code: str) -> None:
        """
        Adds raw HTML code, usage of this method is generally discouraged. Only when there is no other possibility.
        :param html_code: HTML code
        :return: None
        """
        self._doc.asis(html_code)

    def add_text(self, text: str) -> None:
        """
        Adds text to this HTML tag.
        :return: None
        """
        if text is None:
            raise ValueError("Cannot add None as text to the HTML object")
        html = BeautifulSoup(str(text), 'html.parser')
        for part in html.contents:
            if isinstance(part, bs4.element.Tag):
                self._doc.asis(f'<{part.name}>')
                self._doc.text(part.text)
                self._doc.asis(f'</{part.name}>')
            else:
                self._doc.text(str(part))

    def add_module_header(self, text: str, id_: str = None) -> None:
        """
        Adds a sub header.
        :param text: Header text
        :param id_: Html id
        :return: None
        """
        attributes = [('class', 'sub_header')]
        if id_ is not None:
            attributes.append(('id', id_))
        with self.get_tag('div', attributes):
            with self.get_tag('h2'):
                self.add_text(text)

    def add_horizontal_line(self) -> None:
        """
        Adds a horizontal line.
        :return: None
        """
        self._doc.stag('hr')



import logging
import shutil
from pathlib import Path
from typing import List, Union, Tuple, Optional, Sequence

from pacu.app.report.htmlbase import HtmlBase
from pacu.app.report.htmlelement import HtmlElement
from pacu.app.report.htmltablecell import HtmlTableCell


class HtmlReportSection(HtmlElement):
    """
    This class can be used to create a section in the HTML report.
    """

    def __init__(self, title: Optional[str], level: int = 2, subtitle: Optional[str] = None) -> None:
        """
        Initializes a report section.
        :param title: Section title
        :param level: Header level
        :param subtitle: Subtitle to put next to the header
        """
        super(HtmlReportSection, self).__init__('div', attributes=[('class', 'report_section')])
        self._title = title
        if (title is not None) and (subtitle is not None):
            self.add_header_with_subtitle(title, level, subtitle)
        elif title is not None:
            self.add_header(title, level, subtitle)
        self._files = []

    def add_header_with_subtitle(self, header: str, level: int, subtitle: str) -> None:
        """
        Adds a header with a subtitle.
        :param header: Header
        :param level: Header level
        :param subtitle: Subtitle
        :return: None
        """
        with self.get_tag('h{}'.format(str(level))):
            self.add_text(header + ' ')
            with self.get_tag('small', [('class', 'header-subtitle')]):
                self.add_text('({})'.format(subtitle))

    @property
    def files(self) -> List[Path]:
        """
        Returns the files that were added to this report.
        :return: Files
        """
        return self._files

    def copy_files(self, output_directory: Path) -> None:
        """
        Exports the files belonging to this section to the given directory.
        :param output_directory: Output directory
        :return: None
        """
        logging.info("Exporting report section files")
        if output_directory is None or not output_directory.is_dir():
            raise ValueError(f'Invalid output directory: {output_directory}')
        for file_path, relative_path in self.files:
            logging.info(f'Copying file: {file_path} -> {relative_path}')
            if not file_path.is_file():
                raise ValueError(f"Cannot add file (does not exist) '{file_path}'")
            relative_dir = output_directory / relative_path.parent
            if not relative_dir.is_dir():
                relative_dir.mkdir(parents=True)
            shutil.copy(file_path, output_directory / relative_path)

    def add_file(self, input_file: Path, relative_path: Path) -> Path:
        """
        Adds the file to the report.
        Returns the relative path parameter so it is easier to use in list comprehensions.
        :param input_file: Input file
        :param relative_path: path where the file will be saved relative to the report output directory
        :return: Relative path
        """
        logging.info("Adding file to report section: {}".format(relative_path))
        self._files.append((input_file, relative_path,))
        return relative_path

    def add_table(self, data: List[Sequence[Union[str, int, 'HtmlBase']]], column_names: List[str] = None,
                  table_attributes: List[Tuple[str, str]] = None, msg_if_empty: str = 'None found') -> None:
        """
        Adds a table to the report section.
        :param data: Table data
        :param column_names: Column names
        :param table_attributes: Table attributes
        :param msg_if_empty: If set, this message is added when the data is empty
        :return: None
        """
        if (msg_if_empty is not None) and (len(data) == 0) and (column_names is not None):
            data.append([HtmlTableCell(msg_if_empty, attributes=[('colspan', len(column_names))])])
        super().add_table(data, column_names, table_attributes)

    def __repr__(self) -> str:
        """
        Returns the internal representation.
        :return: String representation
        """
        return f"HtmlReportSection(title='{self._title}')"



from dataclasses import dataclass
from pathlib import Path
from typing import List, Dict

import vcf
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
# noinspection PyProtectedMember
from vcf.model import _Record as VCFRecord

from pacu.app.utils.loggingutils import logger


@dataclass(unsafe_hash=True, frozen=True)
class SNPPosition:
    contig: str
    position: int
    reference_base: chr


def parse_vcf_file(path_vcf: Path, include_filtered: bool) -> List[VCFRecord]:
    """
    Parses a single VCF file.
    :param path_vcf: VCF path
    :param include_filtered: If True, filtered variants are included
    :return: List of SNP positions
    """
    vcf_records = []
    logger.debug(f'Parsing VCF file: {path_vcf}')
    for record in list(vcf.Reader(filename=str(path_vcf))):
        # Remove non SNP positions
        if not record.is_snp:
            continue
        # Remove filtered records (unless specified otherwise)
        if (not include_filtered) and len(record.FILTER) > 0:
            continue
        vcf_records.append(record)
    logger.info(f"{path_vcf.name} parsed: {len(vcf_records):,} variant positions")
    return vcf_records


def __get_nucleotides_per_position(paths_vcf: List[Path], names: List[str], include_filtered: bool = False) -> (
        Dict)[SNPPosition, Dict[str, str]]:
    """
    Returns a dictionary with the nucleotide for each sample at each variant position.
    :param paths_vcf: List of input VCF files
    :param include_filtered: If True, filtered positions are retained
    :return: Sample_names, nucleotide per position per sample
    """
    nucl_by_position = {}
    for path_vcf, name in zip(paths_vcf, names):
        for record in parse_vcf_file(path_vcf, include_filtered):
            position = SNPPosition(record.CHROM, record.POS, record.REF)
            if position not in nucl_by_position:
                nucl_by_position[position] = {}
            if len(record.FILTER) > 0:
                nucl_by_position[position][name] = 'N'
            else:
                nucl_by_position[position][name] = str(record.ALT[0])

    # Remove positions with only N
    nucl_by_position = {pos: nucl for pos, nucl in nucl_by_position.items() if not all(
        [x == 'N' for x in nucl.values()])}

    logger.info(f"{len(nucl_by_position):,} SNP positions found across all samples")
    return nucl_by_position


def create_snp_matrix(paths_vcf: List[Path], names: List[str], path_out: Path, include_ref: bool = False) -> None:
    """
    Creates an SNP matrix from the input VCF file.
    :param paths_vcf: List of input VCF files
    :param names: Isolate names
    :param path_out: Output path
    :param include_ref: If True, the reference genome is included
    :return: None
    """
    nucl_by_pos = __get_nucleotides_per_position(paths_vcf, names, False)
    seq_by_sample_name = {name: [] for name in names}
    if include_ref is True:
        seq_by_sample_name['reference'] = []
    for snp_pos, nucl_by_sample in sorted(nucl_by_pos.items(), key=lambda x: x[0].position):
        for name in seq_by_sample_name.keys():
            seq_by_sample_name[name].append(nucl_by_sample.get(name, snp_pos.reference_base))

    # Write output file
    seqs = [SeqRecord(Seq(''.join(seq)), name, description='') for name, seq in seq_by_sample_name.items()]
    with path_out.open('w') as handle:
        SeqIO.write(seqs, handle, 'fasta')



import tempfile
import unittest
from importlib.resources import files
from pathlib import Path

from pacu.app.utils import workflowutils


class TestWorkflowUtils(unittest.TestCase):
    """
    Tests for the workflow utils.
    """

    def test_plot_newick_phylogeny(self) -> None:
        """
        Tests the plot_newick_phylogeny function.
        """
        path_nwk = Path(str(files('pacu').joinpath('resources/testdata/phylogeny.nwk')))
        with tempfile.TemporaryDirectory(prefix='pacu') as dir_:
            # Create visualization
            path_out = Path(dir_, 'tree.png')
            workflowutils.plot_newick_phylogeny(path_nwk, path_out)

            # Verify output file
            self.assertTrue(path_out.exists())
            self.assertGreater(path_out.stat().st_size, 0)

    def test_determine_name_from_fq_ont(self) -> None:
        """
        Tests the 'determine_name_from_fq' function with ONT input.
        :return: None
        """
        self.assertEqual(workflowutils.determine_name_from_fq(fq_ont=Path('/path/to/reads.fastq')), 'reads')
        self.assertEqual(workflowutils.determine_name_from_fq(fq_ont=Path('/path/to/reads.fastq.gz')), 'reads')
        self.assertEqual(workflowutils.determine_name_from_fq(fq_ont=Path('/path/to/reads.fq')), 'reads')
        self.assertEqual(workflowutils.determine_name_from_fq(fq_ont=Path('/path/to/reads.fq.gz')), 'reads')

    def test_determine_name_from_fq_illumina(self) -> None:
        """
        Tests the 'determine_name_from_fq' function with Illumina input.
        :return: None
        """
        self.assertEqual(workflowutils.determine_name_from_fq(
            fq_illumina_1p=Path('/path/to/my-reads_1P.fastq')), 'my-reads')
        self.assertEqual(workflowutils.determine_name_from_fq(
            fq_illumina_1p=Path('/path/to/input_file_1.fastq.gz')), 'input_file')
        self.assertEqual(workflowutils.determine_name_from_fq(
            fq_illumina_1p=Path('/path/to/reads_1.fq')), 'reads')
        self.assertEqual(workflowutils.determine_name_from_fq(
            fq_illumina_1p=Path('/path/to/reads_1P.fq.gz')), 'reads')
        self.assertEqual(workflowutils.determine_name_from_fq(
            fq_illumina_1p=Path('/path/to/reads_S22_L001_R1_001.fastq.gz')), 'reads_S22')
        self.assertEqual(workflowutils.determine_name_from_fq(
            fq_illumina_1p=Path('/path/to/reads_S22_L001_R1_001.fq.gz')), 'reads_S22')



if __name__ == '__main__':
    unittest.main()



import argparse
from importlib.resources import files
from pathlib import Path
from typing import Optional, Sequence

from Bio import SeqIO

from pacu.app.command import Command
from pacu.app.utils import snakemakeutils
from pacu.app.utils.loggingutils import logger
from pacu.version import __version__


class PACU(object):
    """
    Main class to run the PACU pipeline.
    """

    def __init__(self, args: Optional[Sequence[str]] = None) -> None:
        """
        Initializes the main class.
        :param args: Arguments (optional)
        :return: None
        """
        self._args = PACU._parse_arguments(args)
        self._path_html_out = self._args.output.absolute() / self._args.output_html
        if self._path_html_out.exists():
            self._path_html_out.unlink()

    def run(self) -> None:
        """
        Runs the main script.
        :return: None
        """
        logger.info('Starting PACU workflow')
        self._check_dependencies()
        self._validate_input_files()
        path_snakefile = Path(str(files('pacu').joinpath('resources/snp_workflow.smk')))
        if self._args.ref_bed is None:
            path_bed = self._args.dir_working / 'empty_phages.bed'
            path_bed.touch()
            logger.info(f'Creating empty phage BED file: {path_bed}')
            self._args.ref_bed = path_bed
        path_config = self.__create_config_file()
        snakemakeutils.run_snakemake(
            path_snakefile, path_config, [self._path_html_out], self._args.dir_working, self._args.threads)
        logger.info(f'Workflow finished successfully, output available in: {self._args.output}')

    @staticmethod
    def _parse_arguments(args: Optional[Sequence[str]]) -> argparse.Namespace:
        """
        Parses the command line arguments.
        :param args: Arguments
        :return: Parsed arguments.
        """
        parser = argparse.ArgumentParser()

        # Input and output
        parser.add_argument('--ilmn-in', help='Directory with Illumina input BAM files')
        parser.add_argument('--ont-in', help='Directory with ONT input BAM files')
        parser.add_argument('--ref-fasta', required=True, help='Reference FASTA file', type=Path)
        parser.add_argument('--ref-bed', type=Path, help='BED file with phage regions')
        parser.add_argument(
            '--dir-working', type=Path, help='Working directory', default=Path.cwd())
        parser.add_argument('--output', required=True, type=Path, help='Output directory')
        parser.add_argument('--output-html', type=Path, help='Output report name', default='report.html')

        # Parameters
        parser.add_argument(
            '--use-mega', action='store_true',
            help='If set, MEGA is used for the construction of the phylogeny (instead of IQ-TREE)')
        parser.add_argument(
            '--include-ref', action='store_true', help='If set, the reference genome is included in the phylogeny')
        parser.add_argument('--min-snp-af', type=float, default=0.90, help='Minimum allele frequency for variants')
        parser.add_argument('--min-snp-qual', type=int, default=25, help='Minimum SNP quality')
        parser.add_argument('--min-snp-depth', type=int, default=5, help='Minimum SNP depth')
        parser.add_argument('--min-snp-dist', type=int, default=10, help='Minimum distance between SNPs')
        parser.add_argument('--skip-gubbins', action='store_true', help='If set, gubbins is skipped')
        # parser.add_argument('--min-sb-pv', type=float, default=0, help='Minimum P-value for strand bias')
        parser.add_argument(
            '--min-global-depth', type=int, default=5,
            help='Minimum depth for all samples to include positions in SNP analysis')
        parser.add_argument('--min-mq-depth', type=int, default=5, help='MQ cutoff for samtools depth')
        parser.add_argument(
            '--bcftools-filt-af1', action='store_true',
            help='If enabled, allele frequency filtering also considers the VAF value')

        # Image
        parser.add_argument('--image-width', type=int, default=480, help='Image width')
        parser.add_argument('--image-height', type=int, default=480, help='Image height')

        # Threads
        parser.add_argument('--threads', type=int, default=4)
        parser.add_argument(
            '--version', help='Print version and exit', action='version', version=f'PACU {__version__}')

        return parser.parse_args(args)

    def __create_config_file(self) -> Path:
        """
        Creates the config file for the Snakemake workflow.
        :return: Path to config file
        """
        input_dict = {}
        if self._args.ilmn_in is not None:
            for bam_file in sorted(Path(self._args.ilmn_in).glob('*.bam')):
                input_dict[bam_file.stem] = {'bam': str(bam_file.absolute()), 'tech': 'ilmn'}
        if self._args.ont_in is not None:
            for bam_file in sorted(Path(self._args.ont_in).glob('*.bam')):
                input_dict[bam_file.stem] = {'bam': str(bam_file.absolute()), 'tech': 'ont'}
        if len(input_dict) == 0:
            raise FileNotFoundError(
                f'No input datasets found (searched in Illumina={self._args.ilmn_in}, ONT={self._args.ont_in})')
        if len(input_dict) < 4:
            raise ValueError(
                f'At least four input BAM files are required (for bootstrapping), {len(input_dict)} found.')

        config_data = {
            'depth': {
                'min_mq': self._args.min_mq_depth,
                'min_depth': 5
            },
            'filters': {
                'min_af': self._args.min_snp_af,
                'min_depth': self._args.min_snp_depth,
                'min_qual': self._args.min_snp_qual,
                'min_dist': self._args.min_snp_dist
            },
            'image': {
                'width': self._args.image_width,
                'height': self._args.image_height
            },
            'skip_gubbins': self._args.skip_gubbins,
            'include_ref': self._args.include_ref,
            'input': input_dict,
            'phylogeny_method': 'mega' if self._args.use_mega else 'iqtree',
            'output': {
                'html': str(self._path_html_out),
                'dir': str(self._args.output.absolute())
            },
            'reference': {
                'fasta': str(self._args.ref_fasta.absolute()),
                'bed_phages': str(self._args.ref_bed.absolute()),
            }
        }
        return Path(snakemakeutils.generate_config_file(config_data, self._args.dir_working.absolute()))

    def _check_dependencies(self) -> None:
        """
        Checks if the required dependencies are available.
        :return: None
        """
        commands = {
            'bcftools': 'bcftools --version',
            'samtools': 'samtools --version',
            'bedtools': 'bedtools --version',
            'gubbins': 'ml gubbins; gubbins -h',
            'snp-dists': 'snp-dists -v',
            'figtree': 'figtree -help',
        }
        if self._args.skip_gubbins:
            commands.pop('gubbins')
        if self._args.use_mega:
            commands['MEGA'] = 'megacc --version'
        else:
            commands['IQ-TREE'] = 'iqtree2 --version'
        logger.info(f'Checking dependencies')
        for tool, command in commands.items():
            command = Command(command)
            command.run(self._args.dir_working, disable_logging=True)
            if not command.exit_code == 0:
                raise RuntimeError(f"Dependency '{tool}' not available")
            logger.info(f"{tool}: OK")

    def _validate_input_files(self) -> None:
        """
        Checks if the provided input files are valid.
        :return: None
        """
        with open(self._args.ref_fasta) as handle:
            seqs = list(SeqIO.parse(handle, 'fasta'))
            if len(seqs) > 1:
                raise ValueError(f'The reference genome should only contain a single sequence ({len(seqs):,} found)')



import tempfile
import unittest
from importlib.resources import files
from pathlib import Path

from Bio import SeqIO

from pacu.app.utils import snpmatrixutils


class TestSNPMatrixUtils(unittest.TestCase):
    """
    Tests for the SNP matrix utils.
    """

    paths_vcf = [p for p in Path(str(files('pacu').joinpath('resources/testdata/vcfs'))).glob('*.vcf')]

    def test_create_snp_matrix(self) -> None:
        """
        Tests the create_snp_matrix function.
        """
        with tempfile.TemporaryDirectory(prefix='pacu') as dir_:
            # Create SNP matrix
            path_out = Path(dir_, 'snp_matrix.fasta')
            snpmatrixutils.create_snp_matrix(
                TestSNPMatrixUtils.paths_vcf,
                [p.name for p in TestSNPMatrixUtils.paths_vcf],
                path_out)

            # Verify output file
            with path_out.open() as handle:
                seqs = list(SeqIO.parse(handle, 'fasta'))
                self.assertTrue(all(len(s) > 0 for s in seqs))
                self.assertTrue(all(len(s) == len(seqs[0]) for s in seqs))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from pacu import initialize_logging, PACU


def main() -> None:
    """
    Runs the main workflow.
    :return: None
    """
    initialize_logging()
    main_ = PACU()
    main_.run()


if __name__ == '__main__':
    main()



import re
from pathlib import Path
from typing import Dict, Any, List, Union

import yaml

from pacu.app.command import Command
from pacu.app.utils.loggingutils import logger


def generate_config_file(config_data: Dict[str, Any], output_dir: Path, output_basename: str = 'config.yml') -> str:
    """
    Generates a configuration file for Snakemake in YAML file format.
    :param config_data: Configuration data
    :param output_dir: Output directory
    :param output_basename: Output basename
    :return: Path to config file
    """
    config_path = output_dir / output_basename
    if not output_dir.exists():
        output_dir.mkdir(parents=True)
    with config_path.open('w') as handle:
        yaml.dump(config_data, handle)
    logger.info(f"Configuration file created: {config_path}")
    return str(config_path)


def run_snakemake(snakefile: Path, config_path: Path, targets: List[Path], dir_: Path, threads: int = 8) -> Command:
    """
    Helper function to run snakemake workflows.
    :param snakefile: Workflow snakefile
    :param config_path: Path to configuration file
    :param targets: Target output files
    :param dir_: Working directory
    :param threads: Number of threads to use
    :return: Snakemake command
    """
    # Create working directory
    if not dir_.exists():
        logger.debug(f'Creating working directory: {dir_}')
        dir_.mkdir(parents=True)

    # Create and run command
    command = Command(' '.join([
        'snakemake', *[str(x) for x in targets], '--snakefile', str(snakefile), '--configfile', str(config_path),
        '--cores', str(threads)]))
    command.run(dir_)

    # Check if command executed successfully
    if command.exit_code != 0:
        rule_failed = __get_failed_rule(command.stderr)
        logger.error(f"Failed at rule: {rule_failed if rule_failed is not None else 'n/a'}")
        raise RuntimeError(f'Error executing Snakemake')
    return command


def __get_failed_rule(stderr: str) -> Union[str, None]:
    """
    Returns the name of the rule that failed during Snakemake execution.
    :return: Name of the failed rule
    """
    for line in reversed(stderr.splitlines()):
        m = re.match(r'Error in rule (\w+):', line.strip())
        if not m:
            continue
        return m.group(1)



import tempfile
import unittest
from importlib.resources import files
from pathlib import Path

from pacu import PACU


class TestWorkflow(unittest.TestCase):
    """
    Tests for the SNP workflow.
    """

    def test_snp_workflow_iqtree(self) -> None:
        """
        Tests the SNP workflow with mixed input (Illumina + R10) and IQ-TREE tree construction.
        """
        with tempfile.TemporaryDirectory(prefix='pacu') as dir_:
            dir_out = Path(dir_, 'output')
            workflow = PACU([
                '--ref-fasta', str(files('pacu').joinpath('resources/testdata/NC_002695.2-subset.fasta')),
                '--ilmn-in', str(files('pacu').joinpath('resources/testdata/bam/ilmn')),
                '--ont-in', str(files('pacu').joinpath('resources/testdata/bam/ont')),
                '--dir-working', str(dir_),
                '--output', str(dir_out),
                '--threads', '8'
            ])
            workflow.run()

    def test_snp_workflow_mega(self) -> None:
        """
        Tests the SNP workflow with mixed input (Illumina + R10) and MEGA tree construction.
        """
        with tempfile.TemporaryDirectory(prefix='pacu') as dir_:
            dir_out = Path(dir_, 'output')
            workflow = PACU([
                '--ref-fasta', str(files('pacu').joinpath('resources/testdata/NC_002695.2-subset.fasta')),
                '--ilmn-in', str(files('pacu').joinpath('resources/testdata/bam/ilmn')),
                '--ont-in', str(files('pacu').joinpath('resources/testdata/bam/ont')),
                '--use-mega',
                '--dir-working', str(dir_),
                '--output', str(dir_out),
                '--threads', '8'
            ])
            workflow.run()

    def test_snp_workflow_no_gubbins(self) -> None:
        """
        Tests the SNP workflow with mixed input (Illumina + R10) without the 'gubbins' step.
        """
        with tempfile.TemporaryDirectory(prefix='pacu') as dir_:
            dir_ = Path('/scratch/temp/tmp-wf')
            dir_out = Path(dir_, 'output')
            workflow = PACU([
                '--ref-fasta', str(files('pacu').joinpath('resources/testdata/NC_002695.2-subset.fasta')),
                '--ilmn-in', str(files('pacu').joinpath('resources/testdata/bam/ilmn')),
                '--ont-in', str(files('pacu').joinpath('resources/testdata/bam/ont')),
                '--skip-gubbins',
                '--dir-working', str(dir_),
                '--output', str(dir_out),
                '--threads', '8'
            ])
            workflow.run()


if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python
import argparse
from pathlib import Path
from typing import Sequence, Optional, Dict

from pacu import initialize_logging, Command, logger
from pacu.app.utils import workflowutils, trimmingutils, bamutils


class MapToRef(object):
    """
    Main script to map reads to the reference genome.
    """

    def __init__(self, args: Optional[Sequence[str]] = None) -> None:
        """
        Initializes the main script.
        :param args: Arguments (optional)
        :return: None
        """
        self._args = MapToRef._parse_arguments(args)
        self._name = self._determine_name()
        logger.info(f'Input name: {self._name}')

    def run(self) -> None:
        """
        Runs the main script.
        :return: None
        """
        logger.info(f'Starting mapping helper ({self._args.read_type})')
        self._check_dependencies()
        self._rename_galaxy_input()

        # Illumina reads
        if self._args.read_type == 'illumina':
            path_ref = self._illumina_idx_ref()
            if self._args.trim:
                fq_dict = self._illumina_trim()
            else:
                fq_dict = {'1P': self._args.fastq_illumina[0], '2P': self._args.fastq_illumina[1]}
            self._illumina_map(fq_dict, path_ref, self._args.output)

        # ONT reads
        else:
            path_ref = self._ont_idx_ref()
            path_fq = self._ont_trim() if self._args.trim else self._args.fastq_ont
            self._ont_map(path_fq, path_ref, self._args.output)

        # Add a custom tag with the original dataset name (used for Galaxy)
        path_bam_temp = Path(self._args.dir_working, 'tmp.bam')
        self._args.output.rename(path_bam_temp)
        bamutils.add_custom_tag('PACU_name', self._name, path_bam_temp, self._args.output)

    @property
    def ref_name(self) -> str:
        """
        Name of the input reference genome.
        :return: Reference name
        """
        if self._args.ref_fasta_name is None:
            return self._args.ref_fasta.name
        return self._args.ref_fasta_name

    def __symlink_ref_fasta(self, dir_: Path) -> Path:
        """
        Creates a symlink for the reference FASTA file.
        :param dir_: Directory
        :return: Path to symlink
        """
        dir_.mkdir(exist_ok=True, parents=True)
        path_ref_link = dir_ / self.ref_name
        path_ref_link.symlink_to(self._args.ref_fasta)
        logger.debug(f'Creating symlink for reference genome: {path_ref_link} -> {self._args.ref_fasta}')
        return path_ref_link

    def _illumina_trim(self) -> Dict[str, Path]:
        """
        Trims the Illumina reads.
        :return: None
        """
        logger.info(f'Trimming Illumina reads')

        # Create directory
        dir_trim = Path(self._args.dir_working, 'trim')
        dir_trim.mkdir(exist_ok=True, parents=True)

        # Run trimming
        command = Command(' '.join([
            'trimmomatic', 'PE',
            '-baseout', f'{self._name}.fastq.gz',
            f'-threads {self._args.threads}',
            str(self._args.fastq_illumina[0]),
            str(self._args.fastq_illumina[1]),
            f"ILLUMINACLIP:{Path(trimmingutils.trimmomatic_dir_adapters(), 'NexteraPE-PE.fa')}:2:30:10",
            'LEADING:10',
            'TRAILING:10',
            'SLIDINGWINDOW:4:20',
            'MINLEN:40',
            '-phred33'
        ]))
        command.run(dir_trim)
        if not command.exit_code == 0:
            raise RuntimeError(f'Error trimming reads: {command.stderr}')
        return trimmingutils.trimmomatic_collect_output(dir_trim)

    def _illumina_idx_ref(self) -> Path:
        """
        Creates an index for the reference genome for Illumina mapping.
        :return: Path to indexed FASTA file
        """
        dir_idx = self._args.dir_working / 'bt2_idx'
        path_ref_link = self.__symlink_ref_fasta(dir_idx)
        command = Command(' '.join(['bowtie2-build', str(path_ref_link.name), str(path_ref_link.name)]))
        command.run(dir_idx)
        if not command.exit_code == 0:
            raise RuntimeError(f'Error creating Bowtie2 index: {command.stderr}')
        logger.info(f'Bowtie2 index created in: {dir_idx}')
        return path_ref_link

    def _illumina_map(self, fq_dict: Dict[str, Path], path_ref: Path, path_out: Path) -> None:
        """
        Maps the Illumina reads to the reference genome.
        :param fq_dict: Input FASTQ dictionary
        :param path_ref: Path to reference genome
        :param path_out: Path to output BAM file
        :return: None
        """
        # Construct bowtie2 command
        parts_bt2 = [
            'bowtie2', '--end-to-end', '--sensitive', f'-x {path_ref}', f"-1 {fq_dict['1P']}", f"-2 {fq_dict['2P']}"]

        # Add orphaned reads (if available)
        orphaned_reads = [fq_dict.get(k) for k in ('1U', '2U') if fq_dict.get(k) is not None]
        if len(orphaned_reads) > 0:
            parts_bt2.append(f"-U {','.join(str(p) for p in orphaned_reads)}")
        parts_bt2.append(f'-p {self._args.threads}')

        # Execute the command
        command = Command(' '.join([
            *parts_bt2,
            f'| samtools view -b | samtools sort - > {path_out}'
        ]))
        command.run(self._args.dir_working)
        if not command.exit_code == 0:
            raise RuntimeError(f'Error mapping reads: {command.stderr}')

    def _ont_trim(self, min_qual: int = 7, min_len: int = 500) -> Path:
        """
        Trims the input ONT reads.
        :param min_qual: Minimum read quality
        :param min_len: Minimum read length
        :return: Path to trimmed reads
        """
        logger.info(f'Trimming ONT reads')

        # Create directory
        dir_trim = Path(self._args.dir_working, 'trim')
        dir_trim.mkdir(exist_ok=True, parents=True)

        # Run the command
        path_out = dir_trim / f'{self._name}.fastq.gz'
        command = Command(' '.join([
            'seqkit seq',
            '--min-qual', str(min_qual),
            '--min-len', str(min_len),
            str(self._args.fastq_ont),
            f'> {path_out}'
        ]))
        command.run(path_out.parent)
        if not command.exit_code == 0:
            raise RuntimeError(f'Error trimming reads: {command.stderr}')
        return path_out

    def _ont_idx_ref(self) -> Path:
        """
        Creates an index for the reference genome for ONT mapping.
        :return: Path to indexed FASTA file
        """
        dir_idx = self._args.dir_working / 'mm2_idx'
        path_ref_link = self.__symlink_ref_fasta(dir_idx)
        path_mni = path_ref_link.parent / f'{path_ref_link.name}.mni'
        command = Command(f'minimap2 -x map-ont -d {path_mni} {path_ref_link}')
        command.run(dir_idx)
        if not command.exit_code == 0:
            raise RuntimeError(f'Error creating minimap2 index: {command.stderr}')
        logger.info(f'Minimap2 index created in: {dir_idx}')
        return path_mni

    def _ont_map(self, path_fq: Path, path_ref: Path, path_out: Path) -> None:
        """
        Maps the ONT reads
        :param path_fq: Path to the input FASTQ file
        :param path_ref: Path to reference genome
        :param path_out: Path to output BAM file
        :return: None
        """
        command = Command(' '.join([
            'minimap2',
            f'-a {path_ref}',
            str(path_fq),
            f'-t {self._args.threads}',
            f'| samtools view -b | samtools sort - > {path_out}'
        ]))
        command.run(self._args.dir_working)
        if not command.exit_code == 0:
            raise RuntimeError(f'Error mapping reads: {command.stderr}')

    @staticmethod
    def _parse_arguments(args: Optional[Sequence[str]]) -> argparse.Namespace:
        """
        Parses the command line arguments.
        :param args: Arguments
        :return: Parsed arguments.
        """
        parser = argparse.ArgumentParser()
        # FASTQ input
        parser.add_argument('--read-type', choices=['ont', 'illumina'], required=True)
        parser.add_argument('--fastq-ont', type=Path)
        parser.add_argument('--fastq-ont-name', type=str, help='Original FASTQ name (used for Galaxy)')
        parser.add_argument('--fastq-illumina', type=Path, nargs=2)
        parser.add_argument(
            '--fastq-illumina-names', type=str, nargs=2, help='Original FASTQ names (used for Galaxy)')

        # FASTA input
        parser.add_argument('--ref-fasta', required=True, help='Reference FASTA file', type=Path)
        parser.add_argument(
            '--ref-fasta-name',  type=Path, help='Original FASTA file name (used for Galaxy)')

        # Other options
        parser.add_argument(
            '--dir-working', type=Path, help='Working directory', required=True)
        parser.add_argument('--trim', action='store_true', help='Trim reads prior to mapping')
        parser.add_argument('--output', required=True, type=Path, help='Output BAM file')
        parser.add_argument('--threads', type=int, default=4)
        return parser.parse_args(args)

    def _check_dependencies(self) -> None:
        """
        Checks if the required dependencies are available.
        :return: None
        """
        commands = {
            'bowtie2': 'bowtie2 --version',
            'minimap2': 'minimap2 --version',
            'samtools': 'samtools --version',
        }
        if self._args.trim:
            commands['trimmomatic'] = 'trimmomatic -version'
            commands['seqkit'] = 'seqkit version'
        logger.info(f'Checking dependencies')
        for tool, command_str in commands.items():
            command = Command(command_str)
            command.run(self._args.dir_working, disable_logging=True)
            if not command.exit_code == 0:
                raise RuntimeError(f"Dependency '{tool}' not available")
            logger.info(f"{tool}: OK")

    def _rename_galaxy_input(self) -> None:
        """
        Renames the input files when the script is executed through Galaxy.
        :return: None
        """
        if self._args.fastq_illumina_names is not None:
            # Symlink the input files
            Path(self._args.dir_working, 'input').mkdir(parents=True, exist_ok=True)
            for idx in (0, 1):
                is_gzipped = workflowutils.is_gzipped(self._args.fastq_illumina[idx])
                path_link = Path(
                    self._args.dir_working, 'input', f'{self._name}_{idx+1}P.fastq' + ('.gz' if is_gzipped else ''))
                logger.debug(f'Creating link: {path_link} -> {self._args.fastq_illumina[idx]}')
                path_link.absolute().symlink_to(self._args.fastq_illumina[idx].absolute())
                self._args.fastq_illumina[idx] = path_link
        elif self._args.fastq_ont_name is not None:
            Path(self._args.dir_working, 'input').mkdir(parents=True, exist_ok=True)
            is_gzipped = workflowutils.is_gzipped(self._args.fastq_ont)
            path_link = Path(
                self._args.dir_working, 'input', f'{self._name}.fastq' + ('.gz' if is_gzipped else '')).absolute()
            path_link.symlink_to(self._args.fastq_ont.absolute())
        else:
            logger.debug(f'Not renaming inputs')

    def _determine_name(self) -> str:
        """
        Determines the input name.
        :return: Input name
        """
        if self._args.read_type == 'ont':
            if self._args.fastq_ont_name is not None:
                return workflowutils.determine_name_from_fq(fq_ont=Path(self._args.fastq_ont_name))
            return workflowutils.determine_name_from_fq(fq_ont=self._args.fastq_ont)
        else:
            if self._args.fastq_illumina_names is not None:
                return workflowutils.determine_name_from_fq(fq_illumina_1p=Path(self._args.fastq_illumina_names[0]))
            return workflowutils.determine_name_from_fq(fq_illumina_1p=self._args.fastq_illumina[0])

def main() -> None:
    """
    Wrapper around the main script.
    :return: None
    """
    initialize_logging()
    map_to_ref = MapToRef()
    map_to_ref.run()


if __name__ == '__main__':
    main()



import binascii
import re
import tempfile
from importlib.resources import files
from pathlib import Path
from typing import Union

import numpy
import pandas as pd
import vcf
from Bio import Phylo

from pacu.app.command import Command
from pacu.app.utils.loggingutils import logger


def is_new_region(record: pd.Series) -> bool:
    """
    Function to check if a region overlaps the previous one.
    :param record: BED file record
    :return: True if it is a new region (not overlapping the previous one), False otherwise
    """
    if numpy.isnan(record['shift_pos']):
        return False
    elif int(record['pos']) - 1 != int(record['shift_pos']):
        return True
    elif record['chr'] != record['chr']:
        return True
    return False


def calculate_overlaps(size: int, bed_phages: Path, bed_gubbins: Path, bed_depth: Path) -> pd.DataFrame:
    """
    Calculates the size of the overlaps between the BED files.
    :param bed_phages: BED file with phage regions
    :param bed_gubbins: BED file with recombination detected by Gubbins
    :param bed_depth: BED file with low-depth positions
    :param size: Reference genome size
    """
    data_overlap = pd.DataFrame(data={'pos': range(size)})

    bed_dict = {'phages': bed_phages, 'gubbins': bed_gubbins, 'depth': bed_depth}
    for key, path_bed in bed_dict.items():
        data_overlap[key] = False
        data_bed = pd.read_table(path_bed, names=['seq_id', 'start', 'end', 'name'])
        logger.info(f'{len(data_bed)} regions parsed from {path_bed.name}')
        for region in data_bed.itertuples():
            # noinspection PyUnresolvedReferences
            pos = (data_overlap['pos'][(data_overlap['pos'] > region.start) & (data_overlap['pos'] <= region.end)])
            data_overlap.loc[pos, key] = True

    data_overlap = data_overlap[data_overlap[bed_dict.keys()].any(axis=1)]
    logger.info(f'{len(data_overlap):,} positions removed in total')
    return data_overlap


def count_covered_positions(bed_file: Path) -> int:
    """
    Counts the number of genomic positions from a BED file.
    - *Important!*: Only works for BED files with non-overlapping features
    :param bed_file: Input BED file
    :return: Number of covered positions
    """
    data_in = pd.read_table(bed_file, usecols=[0, 1, 2], names=['chr', 'start', 'end'])
    data_in['interval_size'] = data_in['end'] - data_in['start']
    return sum(data_in['interval_size'])


def count_overlap(bed_file_a: Path, bed_file_b: Path) -> int:
    """
    Counts the overlap between two bed files.
    :param bed_file_a: Input bed file A
    :param bed_file_b: Input bed file B
    :return: Number of overlaps
    """
    with tempfile.TemporaryDirectory(prefix='pacu') as dir_temp:
        command = Command(' '.join([
            f'bedtools multiinter -i {bed_file_a} {bed_file_b}',
            '|', "awk -F'\t' 'BEGIN{SUM=0}{SUM+=$3-$2 }END{print SUM}'"
        ]))
        command.run(Path(dir_temp))
    return int(command.stdout)


def calculate_distance(row: pd.Series) -> Union[int, None]:
    """
    Calculates the distance to the closest SNP.
    :param row: Table row
    :return: Distance to SNP
    """
    if pd.isna(row['chr_next']) or pd.isna(row['chr_prev']):
        return None

    # Distance to previous SNP
    dist_to_prev = None
    if row['chr_prev'] == row['chr']:
        dist_to_prev = row['pos'] - row['pos_prev']

    # Distance to next SNP
    dist_to_next = None
    if row['chr_next'] == row['chr']:
        dist_to_next = row['pos_next'] - row['pos']

    # Single variant on chromosome
    if dist_to_next is None and dist_to_prev is None:
        return None

    return min(x for x in (dist_to_prev, dist_to_next) if x is not None)


def filter_snp_distance(vcf_in: Path, vcf_out: Path, min_dist: int = 10) -> None:
    """
    Soft-filters SNPs that are located within less than other SNPs.
    Note that the filter considers all SNPs (including those at filtered positions)
    :param min_dist: Minimum distance between SNPs
    :param vcf_in: Input VCF file
    :param vcf_out: Output VCF file
    :return: None
    """
    # Parse input SNPs
    records_snps = []
    with vcf_in.open() as handle:
        for row in vcf.Reader(handle):
            records_snps.append({'chr': row.CHROM, 'pos': row.POS})
    data_snps = pd.DataFrame(records_snps)
    logger.info(f'{len(data_snps):,} SNPs parsed')

    # Calculate distance to closest SNP
    if len(data_snps) > 0:
        data_snps['pos_next'] = data_snps['pos'].shift(-1)
        data_snps['chr_next'] = data_snps['chr'].shift(-1)
        data_snps['pos_prev'] = data_snps['pos'].shift(1)
        data_snps['chr_prev'] = data_snps['chr'].shift(1)
        data_snps['dist'] = data_snps.apply(lambda x: calculate_distance(x), axis=1)

        # Extract the keys of the variants that need to be filtered
        keys_removed = [(x['chr'], x['pos']) for x in data_snps[data_snps['dist'].apply(
            lambda x: x is not None and x <= min_dist)].to_dict('records')]
    else:
        keys_removed = []
    logger.info(f'Filtering {len(keys_removed):,} variants based on distance')

    # Save output file
    with vcf_in.open() as handle_in, vcf_out.open('w') as handle_out:
        vcf_reader = vcf.Reader(handle_in)
        vcf_writer = vcf.Writer(handle_out, template=vcf_reader)

        # Update filtering records
        for record in vcf_reader:
            if (record.CHROM, record.POS) in keys_removed:
                record.add_filter('distance')

            # Write to output file
            vcf_writer.write_record(record)
        vcf_writer.close()
    logger.info(f'Output VCF file created: {vcf_out}')


def sort_snp_dist_matrix(path_nwk: Path, path_tsv: Path, path_out: Path) -> None:
    """
    Sorts the input SNP distance matrix based on the order of the nodes in the input phylogeny.
    :param path_nwk: Path to the input Newick file
    :param path_tsv: Path to the input TSV file
    :param path_out: Output path
    :return: None
    """
    with path_nwk.open() as handle:
        tree = Phylo.read(handle, 'newick')
        tree.root_at_midpoint()
        node_order = [n.name for n in tree.find_clades() if n.is_terminal()]

    # Parse distance information
    data_in_dist = pd.read_table(path_tsv, index_col=0)

    # Sort rows & columns
    data_in_dist = data_in_dist.loc[node_order]
    data_in_dist = data_in_dist[node_order]
    data_in_dist.to_csv(path_out, sep='\t', index=True, index_label='isolate')


def plot_newick_phylogeny(path_nwk: Path, path_out: Path, width: int = 600, height: int = 600) -> None:
    """
    Plots the input Newick phylogeny in PNG format.
    :param path_nwk: Input Newick file
    :param path_out: Output PNG path
    :param width: Image width
    :param height: Image height
    :return: None
    """
    with tempfile.NamedTemporaryFile(prefix='pacu') as file_:
        # Convert the tree to Nexus format
        Phylo.convert(str(path_nwk), 'newick', file_.name, 'nexus')

        # Add the figtree code
        path_template = Path(str(files('pacu').joinpath('resources/figtree_template.txt')))
        with open(file_.name, 'a') as handle_out, open(path_template) as handle_in:
            handle_out.write('\n')
            for line in handle_in.readlines():
                handle_out.write(line)

        # Create the visualization
        command = Command(' '.join([
            'figtree',
            '-graphic PNG',
            f'-width {width}',
            f'-height {height}',
            str(file_.name),
            str(path_out.absolute())
        ]))
        command.run(path_out.parent)
        if not command.exit_code == 0:
            raise RuntimeError(f"Error visualizing tree: {command.stderr}")
        logger.info(f'Visualization exported to: {path_out}')


def sanitize_bam_input(name: str) -> str:
    """
    Sanitizes the BAM input file name.
    :param name: Name
    :return: None
    """
    invalid_chars = '/!@#$'
    if not name.endswith('.bam'):
        return f'{name}.bam'
    return ''.join(c for c in name if c not in invalid_chars)


PATTERNS_FQ_PE = [
    r'(.*)_1P?\.(fastq|fq)(\.gz)?',
    r'(.*)_L\d+_R1_\d+\.(fastq|fq)(\.gz)?'
]

def determine_name_from_fq(fq_ont: Path = None, fq_illumina_1p: Path = None) -> str:
    """
    Determines the sample name from the FASTQ input.
    :param fq_ont: Input ONT FASTQ file
    :param fq_illumina_1p: Input illumina forward FASTQ file
    :return: Sample name
    """
    if fq_ont is not None:
        return re.sub(r'\.(fastq|fq)(\.gz)?', '', fq_ont.name)
    elif fq_illumina_1p is not None:
        for pattern in PATTERNS_FQ_PE:
            m = re.match(pattern, fq_illumina_1p.name)
            if not m:
                continue
            return m.group(1)
        logger.warning(f'FASTQ input format does not match known formats')
        return re.sub('(fastq|fq)(\.gz)', '', fq_illumina_1p.name)
    else:
        raise ValueError('No FASTQ file provided')


def is_gzipped(path: Path) -> bool:
    """
    Checks if the given file is compressed with gzip.
    :param path: Path
    :return: True if gzipped, False otherwise
    """
    with path.open('rb') as handle:
        magic_number = binascii.hexlify(handle.read(2))
    return magic_number == b'1f8b'



import tempfile
import unittest
from importlib.resources import files
from pathlib import Path

from pacu.app.utils import iqtreeutils


class TestIqTreeUtils(unittest.TestCase):
    """
    Tests for the IQ-TREE utils.
    """

    path_fasta = Path(str(files('pacu').joinpath('resources/testdata/snp_matrix.fasta')))

    def test_tree_construction(self) -> None:
        """
        Tests the model selection functionality with regular input.
        """
        with tempfile.TemporaryDirectory(prefix='pacu') as dir_:
            path_out_nwk = Path(dir_, 'iqtree.nwk')
            command = iqtreeutils.run_ml_tree_construction(
                path_fasta=TestIqTreeUtils.path_fasta,
                path_out=path_out_nwk,
                threads=4
            )

            # Verify that the model selection worked
            model = iqtreeutils.extract_selected_model(command.stdout)
            self.assertIsNotNone(model)

            # Verify that the output file is OK
            self.assertTrue(path_out_nwk.exists())
            self.assertGreater(path_out_nwk.stat().st_size, 0)


if __name__ == '__main__':
    unittest.main()
